@@ -89,6 +89,7 @@ struct _GKeyFileGroup
   const gchar *name;  /* NULL for above first group (which will be comments) */
 
   GKeyFileKeyValuePair *comment; /* Special comment that is stuck to the top of a group */
+  gboolean has_trailing_blank_line;
 
   GList *key_value_pairs; 
 
@@ -105,8 +106,8 @@ struct _GKeyFileKeyValuePair
 };
 
 static gint                  find_file_in_data_dirs            (const gchar            *file,
+								const gchar           **data_dirs,
 								gchar                 **output_file,
-								gchar                ***data_dirs,
 								GError                **error);
 static gboolean              g_key_file_load_from_fd           (GKeyFile               *key_file,
 								gint                    fd,
@@ -196,7 +197,7 @@ g_key_file_error_quark (void)
 static void
 g_key_file_init (GKeyFile *key_file)
 {  
-  key_file->current_group = g_new0 (GKeyFileGroup, 1);
+  key_file->current_group = g_slice_new0 (GKeyFileGroup);
   key_file->groups = g_list_prepend (NULL, key_file->current_group);
   key_file->start_group = NULL;
   key_file->parse_buffer = g_string_sized_new (128);
@@ -228,8 +229,9 @@ g_key_file_clear (GKeyFile *key_file)
 /**
  * g_key_file_new:
  *
- * Creates a new empty #GKeyFile object. Use g_key_file_load_from_file(),
- * g_key_file_load_from_data() or g_key_file_load_from_data_dirs() to
+ * Creates a new empty #GKeyFile object. Use
+ * g_key_file_load_from_file(), g_key_file_load_from_data(),
+ * g_key_file_load_from_dirs() or g_key_file_load_from_data_dirs() to
  * read an existing key file.
  *
  * Return value: an empty #GKeyFile.
@@ -241,7 +243,7 @@ g_key_file_new (void)
 {
   GKeyFile *key_file;
 
-  key_file = g_new0 (GKeyFile, 1);
+  key_file = g_slice_new0 (GKeyFile);
   g_key_file_init (key_file);
 
   return key_file;
@@ -271,16 +273,16 @@ g_key_file_set_list_separator (GKeyFile *key_file,
 /* Iterates through all the directories in *dirs trying to
  * open file.  When it successfully locates and opens a file it
  * returns the file descriptor to the open file.  It also
- * outputs the absolute path of the file in output_file and
- * leaves the unchecked directories in *dirs.
+ * outputs the absolute path of the file in output_file.
  */
 static gint
 find_file_in_data_dirs (const gchar   *file,
+                        const gchar  **dirs,
                         gchar        **output_file,
-                        gchar       ***dirs,
                         GError       **error)
 {
-  gchar **data_dirs, *data_dir, *path;
+  const gchar **data_dirs, *data_dir;
+  gchar *path;
   gint fd;
 
   path = NULL;
@@ -289,7 +291,7 @@ find_file_in_data_dirs (const gchar   *file,
   if (dirs == NULL)
     return fd;
 
-  data_dirs = *dirs;
+  data_dirs = dirs;
 
   while (data_dirs && (data_dir = *data_dirs) && fd < 0)
     {
@@ -332,14 +334,12 @@ find_file_in_data_dirs (const gchar   *file,
       data_dirs++;
     }
 
-  *dirs = data_dirs;
-
   if (fd < 0)
     {
       g_set_error (error, G_KEY_FILE_ERROR,
                    G_KEY_FILE_ERROR_NOT_FOUND,
                    _("Valid key file could not be "
-                     "found in data dirs")); 
+                     "found in search dirs"));
     }
 
   if (output_file != NULL && fd > 0)
@@ -539,6 +539,80 @@ g_key_file_load_from_data (GKeyFile       *key_file,
   return TRUE;
 }
 
+/**
+ * g_key_file_load_from_dirs:
+ * @key_file: an empty #GKeyFile struct
+ * @file: a relative path to a filename to open and parse
+ * @search_dirs: %NULL-terminated array of directories to search
+ * @full_path: return location for a string containing the full path
+ *   of the file, or %NULL
+ * @flags: flags from #GKeyFileFlags
+ * @error: return location for a #GError, or %NULL
+ *
+ * This function looks for a key file named @file in the paths
+ * specified in @search_dirs, loads the file into @key_file and
+ * returns the file's full path in @full_path.  If the file could not
+ * be loaded then an %error is set to either a #GFileError or
+ * #GKeyFileError.
+ *
+ * Return value: %TRUE if a key file could be loaded, %FALSE othewise
+ *
+ * Since: 2.14
+ **/
+gboolean
+g_key_file_load_from_dirs (GKeyFile       *key_file,
+                           const gchar    *file,
+                           const gchar   **search_dirs,
+                           gchar         **full_path,
+                           GKeyFileFlags   flags,
+                           GError        **error)
+{
+  GError *key_file_error = NULL;
+  const gchar **data_dirs;
+  gchar *output_path;
+  gint fd;
+  gboolean found_file;
+
+  g_return_val_if_fail (key_file != NULL, FALSE);
+  g_return_val_if_fail (!g_path_is_absolute (file), FALSE);
+  g_return_val_if_fail (search_dirs != NULL, FALSE);
+
+  found_file = FALSE;
+  data_dirs = search_dirs;
+  output_path = NULL;
+  while (*data_dirs != NULL && !found_file)
+    {
+      g_free (output_path);
+
+      fd = find_file_in_data_dirs (file, data_dirs, &output_path,
+                                   &key_file_error);
+
+      if (fd < 0)
+        {
+          if (key_file_error)
+            g_propagate_error (error, key_file_error);
+ 	  break;
+        }
+
+      found_file = g_key_file_load_from_fd (key_file, fd, flags,
+	                                    &key_file_error);
+      close (fd);
+
+      if (key_file_error)
+        {
+	  g_propagate_error (error, key_file_error);
+	  break;
+        }
+    }
+
+  if (found_file && full_path)
+    *full_path = output_path;
+  else
+    g_free (output_path);
+
+  return found_file;
+}
+
 /**
  * g_key_file_load_from_data_dirs:
  * @key_file: an empty #GKeyFile struct
@@ -564,21 +638,18 @@ g_key_file_load_from_data_dirs (GKeyFile       *key_file,
 				GKeyFileFlags   flags,
 				GError        **error)
 {
-  GError *key_file_error = NULL;
-  gchar **all_data_dirs, **data_dirs;
+  gchar **all_data_dirs;
   const gchar * user_data_dir;
   const gchar * const * system_data_dirs;
   gsize i, j;
-  gchar *output_path;
-  gint fd;
   gboolean found_file;
-  
+
   g_return_val_if_fail (key_file != NULL, FALSE);
   g_return_val_if_fail (!g_path_is_absolute (file), FALSE);
 
   user_data_dir = g_get_user_data_dir ();
   system_data_dirs = g_get_system_data_dirs ();
-  all_data_dirs = g_new0 (gchar *, g_strv_length ((gchar **)system_data_dirs) + 2);
+  all_data_dirs = g_new (gchar *, g_strv_length ((gchar **)system_data_dirs) + 2);
 
   i = 0;
   all_data_dirs[i++] = g_strdup (user_data_dir);
@@ -586,39 +657,14 @@ g_key_file_load_from_data_dirs (GKeyFile       *key_file,
   j = 0;
   while (system_data_dirs[j] != NULL)
     all_data_dirs[i++] = g_strdup (system_data_dirs[j++]);
+  all_data_dirs[i] = NULL;
 
-  found_file = FALSE;
-  data_dirs = all_data_dirs;
-  output_path = NULL;
-  while (*data_dirs != NULL && !found_file)
-    {
-      g_free (output_path);
-
-      fd = find_file_in_data_dirs (file, &output_path, &data_dirs, 
-                                   &key_file_error);
-      
-      if (fd < 0)
-        {
-          if (key_file_error)
-            g_propagate_error (error, key_file_error);
- 	  break;
-        }
-
-      found_file = g_key_file_load_from_fd (key_file, fd, flags,
-	                                    &key_file_error);
-      close (fd);
-      
-      if (key_file_error)
-        {
-	  g_propagate_error (error, key_file_error);
-	  break;
-        }
-    }
-
-  if (found_file && full_path)
-    *full_path = output_path;
-  else 
-    g_free (output_path);
+  found_file = g_key_file_load_from_dirs (key_file,
+                                          file,
+                                          (const gchar **)all_data_dirs,
+                                          full_path,
+                                          flags,
+                                          error);
 
   g_strfreev (all_data_dirs);
 
@@ -639,7 +685,7 @@ g_key_file_free (GKeyFile *key_file)
   g_return_if_fail (key_file != NULL);
   
   g_key_file_clear (key_file);
-  g_free (key_file);
+  g_slice_free (GKeyFile, key_file);
 }
 
 /* If G_KEY_FILE_KEEP_TRANSLATIONS is not set, only returns
@@ -722,13 +768,15 @@ g_key_file_parse_comment (GKeyFile     *key_file,
   
   g_assert (key_file->current_group != NULL);
 
-  pair = g_new0 (GKeyFileKeyValuePair, 1);
-  
+  pair = g_slice_new (GKeyFileKeyValuePair);
   pair->key = NULL;
   pair->value = g_strndup (line, length);
   
   key_file->current_group->key_value_pairs =
     g_list_prepend (key_file->current_group->key_value_pairs, pair);
+
+  if (length == 0 || line[0] != '#')
+    key_file->current_group->has_trailing_blank_line = TRUE;
 }
 
 static void
@@ -959,6 +1007,7 @@ g_key_file_to_data (GKeyFile  *key_file,
 {
   GString *data_string;
   GList *group_node, *key_file_node;
+  gboolean has_blank_line = TRUE;
 
   g_return_val_if_fail (key_file != NULL, NULL);
 
@@ -972,10 +1021,13 @@ g_key_file_to_data (GKeyFile  *key_file,
 
       group = (GKeyFileGroup *) group_node->data;
 
+      /* separate groups by at least an empty line */
+      if (!has_blank_line)
+	g_string_append_c (data_string, '\n');
+      has_blank_line = group->has_trailing_blank_line;
+
       if (group->comment != NULL)
         g_string_append_printf (data_string, "%s\n", group->comment->value);
-      else if (group_node->next) /* separate groups by at least an empty line */
-        g_string_append_c (data_string, '\n');
 
       if (group->name != NULL)
         g_string_append_printf (data_string, "[%s]\n", group->name);
@@ -1055,7 +1107,7 @@ g_key_file_get_keys (GKeyFile     *key_file,
 	num_keys++;
     }
   
-  keys = g_new0 (gchar *, num_keys + 1);
+  keys = g_new (gchar *, num_keys + 1);
 
   i = num_keys - 1;
   for (tmp = group->key_value_pairs; tmp; tmp = tmp->next)
@@ -1132,7 +1184,7 @@ g_key_file_get_groups (GKeyFile *key_file,
    * list) is always the comment group at the top,
    * which we skip
    */
-  groups = g_new0 (gchar *, num_groups);
+  groups = g_new (gchar *, num_groups);
 
   group_node = g_list_last (key_file->groups);
   
@@ -1452,7 +1504,7 @@ g_key_file_get_string_list (GKeyFile     *key_file,
     }
 
   len = g_slist_length (pieces);
-  values = g_new0 (gchar *, len + 1); 
+  values = g_new (gchar *, len + 1);
   for (p = pieces, i = 0; p; p = p->next)
     values[i++] = p->data;
   values[len] = NULL;
@@ -1594,7 +1646,7 @@ g_key_file_get_locale_string (GKeyFile     *key_file,
 
       list = _g_compute_locale_variants (locale);
 
-      languages = g_new0 (gchar *, g_slist_length (list) + 1);
+      languages = g_new (gchar *, g_slist_length (list) + 1);
       for (l = list, i = 0; l; l = l->next, i++)
 	languages[i] = l->data;
       languages[i] = NULL;
@@ -1767,13 +1819,14 @@ g_key_file_set_locale_string_list (GKeyFile            *key_file,
  * Returns the value associated with @key under @group_name as a
  * boolean. 
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the value associated with @key cannot be interpreted as a boolean
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then %FALSE is returned and @error is set
+ * to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the value
+ * associated with @key cannot be interpreted as a boolean then %FALSE
+ * is returned and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
+ *
+ * Return value: the value associated with the key as a boolean, or
+ * %FALSE if the key was not found or could not be parsed.
  *
- * Return value: the value associated with the key as a boolean
  * Since: 2.6
  **/
 gboolean
@@ -1860,13 +1913,13 @@ g_key_file_set_boolean (GKeyFile    *key_file,
  * Returns the values associated with @key under @group_name as
  * booleans. If @group_name is %NULL, the start_group is used.
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the values associated with @key cannot be interpreted as booleans
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then %NULL is returned and @error is set to
+ * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the values associated
+ * with @key cannot be interpreted as booleans then %NULL is returned
+ * and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
  *
- * Return value: the values associated with the key as a boolean
+ * Return value: the values associated with the key as a list of
+ * booleans, or %NULL if the key was not found or could not be parsed.
  * 
  * Since: 2.6
  **/
@@ -1897,7 +1950,7 @@ g_key_file_get_boolean_list (GKeyFile     *key_file,
   if (!values)
     return NULL;
 
-  bool_values = g_new0 (gboolean, num_bools);
+  bool_values = g_new (gboolean, num_bools);
 
   for (i = 0; i < num_bools; i++)
     {
@@ -1976,13 +2029,13 @@ g_key_file_set_boolean_list (GKeyFile    *key_file,
  * Returns the value associated with @key under @group_name as an
  * integer. If @group_name is %NULL, the start_group is used.
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the value associated with @key cannot be interpreted as an integer
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then 0 is returned and @error is set to
+ * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the value associated
+ * with @key cannot be interpreted as an integer then 0 is returned
+ * and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
  *
- * Return value: the value associated with the key as an integer.
+ * Return value: the value associated with the key as an integer, or
+ * 0 if the key was not found or could not be parsed.
  *
  * Since: 2.6
  **/
@@ -2072,13 +2125,13 @@ g_key_file_set_integer (GKeyFile    *key_file,
  * Returns the values associated with @key under @group_name as
  * integers. 
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the values associated with @key cannot be interpreted as integers
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then %NULL is returned and @error is set to
+ * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the values associated
+ * with @key cannot be interpreted as integers then %NULL is returned
+ * and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
  *
- * Return value: the values associated with the key as a integer
+ * Return value: the values associated with the key as a list of
+ * integers, or %NULL if the key was not found or could not be parsed.
  *
  * Since: 2.6
  **/
@@ -2107,7 +2160,7 @@ g_key_file_get_integer_list (GKeyFile     *key_file,
   if (!values)
     return NULL;
 
-  int_values = g_new0 (gint, num_ints);
+  int_values = g_new (gint, num_ints);
 
   for (i = 0; i < num_ints; i++)
     {
@@ -2182,16 +2235,16 @@ g_key_file_set_integer_list (GKeyFile     *key_file,
  * @key: a key
  * @error: return location for a #GError
  *
- * Returns the value associated with @key under @group_name as an
- * integer. If @group_name is %NULL, the start_group is used.
+ * Returns the value associated with @key under @group_name as a
+ * double. If @group_name is %NULL, the start_group is used.
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the value associated with @key cannot be interpreted as a double
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then 0.0 is returned and @error is set to
+ * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the value associated
+ * with @key cannot be interpreted as a double then 0.0 is returned
+ * and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
  *
- * Return value: the value associated with the key as a double.
+ * Return value: the value associated with the key as a double, or
+ * 0.0 if the key was not found or could not be parsed.
  *
  * Since: 2.12
  **/
@@ -2281,13 +2334,13 @@ g_key_file_set_double  (GKeyFile    *key_file,
  * Returns the values associated with @key under @group_name as
  * doubles. If @group_name is %NULL, the start group is used.
  *
- * If @key cannot be found then the return value is undefined and
- * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
- * the values associated with @key cannot be interpreted as doubles
- * then the return value is also undefined and @error is set to
- * #G_KEY_FILE_ERROR_INVALID_VALUE.
+ * If @key cannot be found then %NULL is returned and @error is set to
+ * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if the values associated
+ * with @key cannot be interpreted as doubles then %NULL is returned
+ * and @error is set to #G_KEY_FILE_ERROR_INVALID_VALUE.
  *
- * Return value: the values associated with the key as a double
+ * Return value: the values associated with the key as a list of
+ * doubles, or %NULL if the key was not found or could not be parsed.
  *
  * Since: 2.12
  **/
@@ -2316,7 +2369,7 @@ g_key_file_get_double_list  (GKeyFile     *key_file,
   if (!values)
     return NULL;
 
-  double_values = g_new0 (gdouble, num_doubles);
+  double_values = g_new (gdouble, num_doubles);
 
   for (i = 0; i < num_doubles; i++)
     {
@@ -2443,8 +2496,7 @@ g_key_file_set_key_comment (GKeyFile             *key_file,
 
   /* Now we can add our new comment
    */
-  pair = g_new0 (GKeyFileKeyValuePair, 1);
-  
+  pair = g_slice_new (GKeyFileKeyValuePair);
   pair->key = NULL;
   pair->value = g_key_file_parse_comment_as_value (key_file, comment);
   
@@ -2485,8 +2537,7 @@ g_key_file_set_group_comment (GKeyFile             *key_file,
 
   /* Now we can add our new comment
    */
-  group->comment = g_new0 (GKeyFileKeyValuePair, 1);
-  
+  group->comment = g_slice_new (GKeyFileKeyValuePair);
   group->comment->key = NULL;
   group->comment->value = g_key_file_parse_comment_as_value (key_file, comment);
 }
@@ -2523,8 +2574,7 @@ g_key_file_set_top_comment (GKeyFile             *key_file,
   if (comment == NULL)
      return;
 
-  pair = g_new0 (GKeyFileKeyValuePair, 1);
-  
+  pair = g_slice_new (GKeyFileKeyValuePair);
   pair->key = NULL;
   pair->value = g_key_file_parse_comment_as_value (key_file, comment);
   
@@ -2902,7 +2952,7 @@ g_key_file_add_group (GKeyFile    *key_file,
       return;
     }
 
-  group = g_new0 (GKeyFileGroup, 1);
+  group = g_slice_new0 (GKeyFileGroup);
   group->name = g_strdup (group_name);
   group->lookup_map = g_hash_table_new (g_str_hash, g_str_equal);
   key_file->groups = g_list_prepend (key_file->groups, group);
@@ -2920,7 +2970,7 @@ g_key_file_key_value_pair_free (GKeyFileKeyValuePair *pair)
     {
       g_free (pair->key);
       g_free (pair->value);
-      g_free (pair);
+      g_slice_free (GKeyFileKeyValuePair, pair);
     }
 }
 
@@ -3024,7 +3074,7 @@ g_key_file_remove_group_node (GKeyFile *key_file,
     }
 
   g_free ((gchar *) group->name);
-  g_free (group);
+  g_slice_free (GKeyFileGroup, group);
   g_list_free_1 (group_node);
 }
 
@@ -3071,13 +3121,13 @@ g_key_file_add_key (GKeyFile      *key_file,
 {
   GKeyFileKeyValuePair *pair;
 
-  pair = g_new0 (GKeyFileKeyValuePair, 1);
-
+  pair = g_slice_new (GKeyFileKeyValuePair);
   pair->key = g_strdup (key);
   pair->value = g_strdup (value);
 
   g_hash_table_replace (group->lookup_map, pair->key, pair);
   group->key_value_pairs = g_list_prepend (group->key_value_pairs, pair);
+  group->has_trailing_blank_line = FALSE;
   key_file->approximate_size += strlen (key) + strlen (value) + 2;
 }
 
@@ -3327,7 +3377,7 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
 {
   gchar *string_value, *p, *q0, *q;
 
-  string_value = g_new0 (gchar, strlen (value) + 1);
+  string_value = g_new (gchar, strlen (value) + 1);
 
   p = (gchar *) value;
   q0 = q = string_value;
@@ -3433,7 +3483,7 @@ g_key_file_parse_string_as_value (GKeyFile    *key_file,
   /* Worst case would be that every character needs to be escaped.
    * In other words every character turns to two characters
    */
-  value = g_new0 (gchar, 2 * length);
+  value = g_new (gchar, 2 * length);
 
   p = (gchar *) string;
   q = value;
@@ -3589,13 +3639,10 @@ g_key_file_parse_value_as_boolean (GKeyFile     *key_file,
 {
   gchar *value_utf8;
 
-  if (value)
-    {
-      if (strcmp (value, "true") == 0 || strcmp (value, "1") == 0)
-        return TRUE;
-      else if (strcmp (value, "false") == 0 || strcmp (value, "0") == 0)
-        return FALSE;
-    }
+  if (strcmp (value, "true") == 0 || strcmp (value, "1") == 0)
+    return TRUE;
+  else if (strcmp (value, "false") == 0 || strcmp (value, "0") == 0)
+    return FALSE;
 
   value_utf8 = _g_utf8_make_valid (value);
   g_set_error (error, G_KEY_FILE_ERROR,
